/*
 * Copyright (c) 2009 Michael Widerkrantz, mc at the domain hack.org.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/sctp.h>
#include <sys/time.h>
#include <sys/select.h>

#define DEBUG 0
#define SCTP 1

#define FALSE (1 == 0)
#define PORT 4711
#define BSIZE 1024
#define MAXCLIENT 128

#if DEBUG
#define PDEBUG(Args...) \
  fprintf(stderr, "server: "); fprintf(stderr, ##Args);
#else
#define PDEBUG(Args...)
#endif

#define MAX(a, b) (((a) > (b))?(a):(b))

struct client
{
    int sock;
    struct sockaddr_in addr;
};

/* Global variables. */
struct client client[MAXCLIENT];

/* Functions. */
static void zeroclients(void);
static void closeclient(int i);

static void zeroclients(void)
{
    int i;
    
    for (i = 0; i < MAXCLIENT; i ++)
    {
        client[i].sock = -1;
    }
}

/*
 * Send the message in senbuf of length sendbuflen to all clients,
 * possibly excluding the connection identified by idx, if set to
 * anything above zero.
 */
void sendall(char *sendbuf, int sendbuflen, int idx)
{
    int i;
    
    for (i = 0; i < MAXCLIENT; i ++)
    {
        if (-1 == client[i].sock || i == idx)
            continue;

        if (-1 == write(client[i].sock, sendbuf, sendbuflen))
        {
            perror("write");
        }
    } /* for */
}

static void closeclient(int i)
{
    close(client[i].sock);
    client[i].sock = -1;
}

int main(void)
{
    int listensock;
    struct sockaddr_in listenaddr;
    int len;
    fd_set inset;
    int maxfd;
    socklen_t addrsize;
    int done = FALSE;
    u_int8_t buf[BSIZE];
    char sendbuf[BSIZE];
    size_t sendbuflen;
    
    zeroclients();
    
#if SCTP
    listensock = socket(AF_INET, SOCK_STREAM, IPPROTO_SCTP);
#else
    listensock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
#endif
    if (-1 == listensock)
    {
        perror("socket");
        exit(1);
    }

    /*
     * We want to be able to reuse the port directly, should the
     * server die.
     */
    len = 1;
    if (0 != setsockopt(listensock, SOL_SOCKET, SO_REUSEADDR, 
                   (char *)&len, sizeof (len))) 
    {
        perror("setsockopt");
        exit(1);
    }

    listenaddr.sin_family = AF_INET;
    listenaddr.sin_addr.s_addr = INADDR_ANY;
    listenaddr.sin_port = ntohs(PORT);

    if (0 != bind(listensock, (struct sockaddr *)&listenaddr,
                  sizeof (struct sockaddr_in)))
    {
        perror("bind");
        exit(1);
    }
    
    if (0 != listen(listensock, SOMAXCONN))
    {
        perror("listen");
        exit(1);
    }

    while (!done)
    {
        int found;
        int i;
        
        FD_ZERO(&inset);

        FD_SET(listensock, &inset);
        maxfd = listensock;
        
        for (i = 0; i < MAXCLIENT; i ++)
        {
            if (-1 != client[i].sock)
            {
                FD_SET(client[i].sock, &inset);
                maxfd = MAX(maxfd, client[i].sock);
            }
        }

        found = select(maxfd + 1, &inset, NULL, NULL, NULL);
        if (-1 == found)
        {
            perror("select");
            exit(1);
        }

        if (FD_ISSET(listensock, &inset))
        {
            PDEBUG("A new connection!\n");

            /* Find a free slot for this connection. */
            for (i = 0; i < MAXCLIENT; i ++)
            {
                if (-1 == client[i].sock)
                {
                    PDEBUG("New connection is %d.\n", i);
                    addrsize = sizeof (struct sockaddr_in);
                    client[i].sock = accept(listensock,
                                            (struct sockaddr *)&client[i].addr,
                                            &addrsize);
                    if (0 > client[i].sock)
                    {
                        perror("accept");
                        exit(1);
                    }

                    /* Say hello. */
                    snprintf(sendbuf, BSIZE, "Welcome! You are #%d.\n", i);
                    sendbuflen = strlen(sendbuf);
                    if (-1 == write(client[i].sock, &sendbuf, sendbuflen))
                    {
                        perror("write");
                    }

                    /* Tell the others... */
                    snprintf(sendbuf, BSIZE, "New user #%d connected.\n", i);
                    sendbuflen = strlen(sendbuf);
                    sendall(sendbuf, sendbuflen, i);
   
                    /* We found or free spot. Jump out of the loop. */
                    break;
                }
            } /* for */
        } /* ISSET listensock */

        /* Go through all active connections and check for pending data. */
        for (i = 0; i < MAXCLIENT; i ++)
        {
            if (-1 == client[i].sock)
                continue;
            
            if (FD_ISSET(client[i].sock, &inset))
            {
                int bytes;
            
                bytes = read(client[i].sock, &buf, BSIZE);
                if (0 > bytes)
                {
                    closeclient(i);
                    perror("read");
                }
                else if (0 == bytes)
                {
                    PDEBUG("Connection %d (id %s) closed.\n", i, client[i].id);
                    closeclient(i);
                    
                    snprintf(sendbuf, BSIZE, "Client %d disconnected.\n", i);
                    sendbuflen = strlen(sendbuf);
                    sendall(sendbuf, sendbuflen, -1);
                }
                else
                {
#if DEBUG
                    PDEBUG("Received %d bytes from %d\n", bytes, i);
#endif

                    /* Send the received message to all clients. */
                    snprintf(sendbuf, BSIZE, "%d: ", i);
                    sendbuflen = strlen(sendbuf);
                    snprintf(sendbuf, BSIZE, "%d: %s", i, buf);
                    sendbuflen = sendbuflen + bytes;

                    sendall(sendbuf, sendbuflen, -1);
                }
            } /* ISSET client */
        } /* for */
    } /* while */
    
    exit(0);
}
